"""Basic template tags library
"""
import collections
from functools import partial, reduce
import itertools
from .template import Template
from .tag import tag_manager, parse_token


def get_value(var_name, context):
    return context[var_name] if var_name in context else None


def resolve(var_name, context):
    if '.' in var_name:
        fields = var_name.split('.')
        return reduce(Template.get_field,
                      [get_value(fields[0], context)] + fields[1:])
    return get_value(var_name, context)


def exec_with_context(func, context={}, context_diff={}):
    old_values = dict([(var_name,
                        context[var_name] if var_name in context else None)
                       for var_name in context_diff])
    context.update(context_diff)
    result = func(context)
    context.update(old_values)
    return result


def exec_block(block, context):
    return "".join([command(context) for command in block])


def block(token, block, template, loader):
    """Block tag
    """
    # Create inner template for blocks
    tmpl = Template(loader)
    tmpl.commands = block

    # Add template block into list
    if not hasattr(template, 'blocks'):
        template.blocks = {}
    is_new = token not in template.blocks
    template.blocks[token] = tmpl

    # Add function that executes inner template into commands
    if is_new:
        return template.blocks[token]
    else:
        index = template.commands.index(template.parent.blocks[token])
        template.commands[index] = tmpl
        return lambda context: ''

tag_manager.register(
        name='block',
        tag=block,
        is_block_tag=True,
        context_required=False,
        template_required=True,
        loader_required=True,
        is_lazy_tag=False
)


def extend(token, template, loader):
    """Tag used to create tamplates hierarhy
    """
    tokens = parse_token(token)[0]
    template.parent = loader.get_template(tokens[0])
    if not hasattr(template, 'blocks'):
        template.blocks = template.parent.blocks.copy()
    else:
        template.blocks.update(template.parent.blocks)
    template.commands.extend(template.parent.commands)
    return None

tag_manager.register(
        name='extend',
        tag=extend,
        template_required=True,
        loader_required=True,
        is_lazy_tag=False
)


def spaceless(token, block, context):
    """This tag removes unused spaces

    Template
        
        {% spaceless %}
            Some
                    text
        {% endspaceless %}

    will be rendered to:
        
        Some text

    """
    results = [command(context).split('\n') for command in block]
    return "".join([line.lstrip() for line in itertools.chain(*results)])

tag_manager.register(
        name='spaceless',
        tag=spaceless,
        is_block_tag=True,
        context_required=True,
        template_required=False,
        loader_required=False,
        is_lazy_tag=True
)


def with_tag(token, block, context):
    """With tag can be used to set the shorter name for variable used few times

    Example:

        {% with request.context.user.name as user_name %}
            <h1>{{ user_name }}'s profile</h1>
            <span>Hello, {{ user_name }}</span>
            <form action="update_profile" method="post">
                <label>Your name:</label>
                <input type="text" name="user_name" value="{{ user_name }}" />
                <input type="submit" value="Update profile" />
            </form>
        {% endwith %}
    """
    data_field, _, var_name = token.split(' ')
    value = resolve(data_field, context)
    return exec_with_context(partial(exec_block, block), 
                               context, {var_name: value})

tag_manager.register(
        name='with',
        tag=with_tag,
        is_block_tag=True,
        context_required=True,
        template_required=False,
        loader_required=False,
        is_lazy_tag=True
)


def if_tag(token, block, context):
    """If tag can brings some logic into template.

    Example:

        {% if user.is_authenticated %}Hello, {{ user.name }}!{% endif %}

    TODO:
        
        - add else
        - add conditions
    """
    if resolve(token, context):
        return exec_block(block, context)
    return ''

tag_manager.register(
        name='if',
        tag=if_tag,
        is_block_tag=True,
        context_required=True,
        template_required=False,
        loader_required=False,
        is_lazy_tag=True
)


class Forloop:
    '''Class for executing block in loop with a context update
    '''

    def __init__(self, var_name, values, block):
        self.var_name = var_name
        self.values = values
        self.block = block

    @property
    def total(self):
        return len(self.values)

    @property
    def last(self):
        return not self.counter0 < self.total
    
    @property
    def first(self):
        return self.counter0 == 0

    @property
    def counter(self):
        return self.counter0 + 1

    def next(self, context):
        self.counter0 = 0
        for v in self.values:
            context[self.var_name] = v
            yield exec_block(self.block, context)

    def __call__(self, context):
        return "".join([next for next in self.next(context)])



def for_tag(token, block, context):
    """For tag

    Example:

        {% for a in items %}{{ a }}{% endfor %}

    returns for items = [1, 2, 3]:
        
        123

    Also forloop variable will be added into scope. It contains few flags can
    be used to render customized templates

        {% for a in items %}
            {% spaceless %}<span
                    {% if forloop.first %} class="first"{% endif %}
                    {% if forloop.last %} class="last"{% endif %}>
                {{ forloop.counter0 }}. 
                {{ forloop.counter }} from {{ forloop.total }}
            </span>{% endspaceless %}
        {% endfor %}

    returns

        <span class="first">0. 1 from 3</span>
        <span>1. 2 from 3</span>
        <span class="last">2. 3 from 3</span>

    """
    var_name, _, data_field = token.split(' ')
    values = resolve(data_field, context)
    # Check values
    if not isinstance(values, collections.Iterable):
        raise ValueError('%s: "%s" is not iterable' % (data_field, values))

    forloop = Forloop(var_name, values, block)
    return exec_with_context(forloop, context, {'forloop': forloop})

tag_manager.register(
        name='for',
        tag=for_tag,
        is_block_tag=True,
        context_required=True,
        template_required=False,
        loader_required=False,
        is_lazy_tag=True
)



"""Package provides template tags manager and base tags list 
"""
VARIABLE    = 0
STRING      = 1
NUMBER      = 2


def parse_token(token):
    tokens = []
    token_types = []
    delim = None
    sentence = None
    for word in [word for word in token.split(' ') if len(word) > 0]:
        if delim is None:
            idx = word.find('"')
            if idx < 0:
                idx = word.find("'")
            if idx >= 0:
                delim = word[idx]
                if idx > 0:
                    token_types.append(STRING)
                    tokens.append(word[0:idx])
                word = word[idx+1:]
                if delim in word:
                    parts = word.split(delim)
                    tokens.append(parts[0])
                    token_types.append(STRING)
                    if len(parts) > 1 and len(parts[1]) > 0:
                        token_types.append(VARIABLE)
                        tokens.append(parts[1])
                    delim = None
                else:
                    sentence = [word]
            else:
                tokens.append(word)
                try:
                    float(word)
                except:
                    token_types.append(VARIABLE)
                else:
                    token_types.append(NUMBER)
        else:
            if delim in word:
                parts = word.split(delim)
                sentence.append(parts[0])
                tokens.append(" ".join(sentence))
                token_types.append(STRING)
                if len(parts) > 1 and len(parts[1]) > 0:
                    token_types.append(STRING)
                    tokens.append(parts[1])
                delim = None
            else:
                sentence.append(word)
    return tokens, token_types


class TagManager(object):
    """Class used for tags manipulation 
    """

    def __init__(self):
        """Create new tag managet instance 
        """
        super(TagManager, self).__init__()
        self.tags = {}

    def register(self, name, tag, is_block_tag=False, context_required=False,
                 template_required=False, loader_required=False, 
                 is_lazy_tag=True):
        """Register new tag
        """
        self.tags[name] = (
            tag,
            is_block_tag,
            context_required,
            template_required,
            loader_required,
            is_lazy_tag
        )

    def is_tag_exists(self, name):
        """Check is tag exists
        """
        if name not in self.tags:
            raise Exception("Tag '%s' is not registered" % name)
        return self.tags[name]

    def is_block_tag(self, name):
        """Check is tag with specified name is block tag
        """
        return self.is_tag_exists(name)[1]

    def is_lazy_tag(self, name):
        """Check is tag with specified name is lazy tag
        Lazy tag means that it would be executed on template execution. But
        some tags required to be executed on template parsing time.
        """
        return self.is_tag_exists(name)[5]

    def execute(self, name, token, context, block, template, loader):
        """Execute tag 
        """
        tag = self.is_tag_exists(name)
        args = {'token': token}
        if tag[1]:
            args['block'] = block
        if tag[2]:
            args['context'] = context
        if tag[3]:
            args['template'] = template
        if tag[4]:
            args['loader'] = loader
        return tag[0](**args)

tag_manager = TagManager()



'''Package contains default template tags
'''
from decimal import Decimal, ROUND_DOWN, ROUND_HALF_UP
from functools import reduce
from operator import itemgetter
import random as random_module

from .filter import filter_manager


# Numbers

def sum(*args):
    '''Calculate the sum of all the values passed as args and
    '''
    return reduce(lambda x, y: x + float(y), args)
filter_manager.register(sum)

def float_format_args_parse(func, raw_value, format):
    # Parse arguments
    try:
        digits  = abs(int(format))
    except:
        raise Exception('%s arguments error: format is not integer' % func)
    try:
        value   = Decimal(raw_value)
    except:
        raise Exception('%s supports only number values' % func)
    return value, digits

def do_float_format(value, digits, rounding):
    # Make formater
    digit = 0
    formatter = Decimal('1')
    while digit < digits:
        digit += 1
        formatter *= Decimal('0.1')
    return value.quantize(formatter, rounding=rounding)

def floatformat(raw_value, format='0'):
    '''Make pretty float representation

    Lets:
        a = '12.4'
    Then:
        >>> print floatformat(a)
        12
        >>> print floatformat(a, '2')
        12.40
        >>> print floatformat(a, '-2')
        12.4
    '''
    value, digits = float_format_args_parse('floatformat', raw_value, format)
    result = do_float_format(value.copy_abs(), digits, ROUND_DOWN)
    result = str(result.copy_sign(value))
    if format[0] == '-':
        return result.rstrip('0')
    return result
filter_manager.register(floatformat)

def floatround(raw_value, format="0"):
    '''Round a float value according to math rules

    Lets:
        a = '12.45'
    Then:
        >>> print floatround(a)
        12
        >>> print floatround(a, '1')
        12.5
    '''
    value, digits = float_format_args_parse('floatround', raw_value, format)
    return do_float_format(value, digits, ROUND_HALF_UP)
filter_manager.register(floatround)


# Strings

def addslashes(value):
    '''Add a slashes to string
    '''
    return value.replace('\\', '\\\\').replace('"', '\\"').replace("'", "\\'")
filter_manager.register(addslashes)

def capfirst(value):
    '''Capitalizes the first character in string
    '''
    return value and value[0].upper() + value[1:]
filter_manager.register(capfirst)

def stringformat(value, format):
    """Formats the variable according to the format, a string formatting 
    specifier.
        
    This specifier uses Python string formating syntax, with the exception that
    the leading "%" is dropped.

    See http://docs.python.org/lib/typesseq-strings.html for documentation
    of Python string formatting
    """
    return ("%" + str(format)) % value
filter_manager.register(stringformat)

def upper(value):
    '''Convert to upper case
    '''
    return str(value).upper()

def lower(value):
    '''Convert to lower case
    '''
    return str(value).lower()


# Lists, dicts, strings

def dictsort(value, key, order=''):
    '''Sort dict
    '''
    return sorted(value, key=itemgetter(key), reverse=(order != ''))
filter_manager.register(dictsort)

def get(value, index):
    '''Get item with specified index
    '''
    if issubclass(value.__class__, dict):
        return value[sorted(value.keys())[index]]
    return value[index]
filter_manager.register(get)

def first(value):
    '''Get first item from list
    '''
    return get(value, 0)
filter_manager.register(first)

def join(value, joiner):
    '''Join list or items with joiner

    >>> join([1, 2, 3], ' ')
    '1 2 3'
    '''
    return joiner.join([str(item) for item in value])
filter_manager.register(join)

def last(value):
    '''Get last item from list
    '''
    return get(value, len(value) - 1)
filter_manager.register(last)

def length(value):
    '''Return's the length of the string, dict or list
    '''
    return len(value)

def random(value):
    '''Get random item from list or dict
    '''
    return get(value, random_module.random(len(value)))
filter_manager.register(random)

def sort(value, order=''):
    '''Sort list
    '''
    return sorted(value, reverse=(order != ''))
filter_manager.register(sort)



""" Package provides template class """

from collections import deque
from functools import reduce
from decimal import Decimal
try:
    import cStringIO
    StringIO = cStringIO.StringIO
except:
    try:
        import StringIO as sio
        StringIO = sio.StringIO
    except:
        import io
        StringIO = io.StringIO


from .loaders import TemplateLoader
from .filter import filter_manager
from .tag import tag_manager, parse_token


class Template(object):
    """Class represents template
    """
    TEXT = 1
    TOKEN = 2
    ECHO = 3
    FILTER = 4
    TAG = 5
    STRING = 6
    CLOSE = 7

    def __init__(self, loader=TemplateLoader(), name="unnamed"):
        """Create new template instance
        """
        super(Template, self).__init__()
        self.loader = loader
        self.name = name
        self.commands = []
        self.context = {}
        self.loader.register(name, self)

    @staticmethod
    def get_field(obj, field):
        if hasattr(obj, field):
            return getattr(obj, field)
        elif hasattr(obj, '__getitem__') and hasattr(obj, '__contains__'):
            return obj[field]
        raise Exception('Could not get %s from %s' % (field, obj))

    @staticmethod
    def variable(name):
        fields = name.split('.')
        if len(fields) > 1:
            def print_value(context):
                return str(reduce(Template.get_field,
                                  [context[fields[0]]] + fields[1:]))
        else:
            def print_value(context):
                return str(context[name])
        return print_value

    @staticmethod
    def constant(value):
        def print_constant(context):
            return value
        return print_constant

    @staticmethod
    def filter(value):
        '''Parse the tamplte filter
        '''
        parts = value.split('|')
        filters = []
        variable = parts[0]
        for token in parts[1:]:
            if ':' in token:
                parsed = token.split(':')
                if len(parsed) > 1:
                    filter, args_token = parsed
                    args, types = parse_token(args_token)
                else:
                    filter = parsed
                    args, types = (), ()
            else:
                filter = token
                args, types = (), ()
            filters.append((filter, args, types))

        def apply_filters(context):
            def apply_filter(value, pair):
                filter, args, types = pair
                return filter_manager.apply(filter, value, args, types,
                                            context)
            if variable[0] == '"' or variable[0] == "'":
                if variable[0] == variable[-1]:
                    filters.insert(0, variable[1:-1])
                else:
                    raise Exception('Template filter syntax error')
            else:
                try:
                    filters.insert(0, Decimal(variable))
                except:
                    filters.insert(0, Template.variable(variable)(context))
            return str(reduce(apply_filter, filters))
        return apply_filters

    def tag(self, name, token, block):
        if tag_manager.is_lazy_tag(name):
            def execute_tag(context):
                return tag_manager.execute(name, token, context, block, self,
                                           self.loader)
            return execute_tag
        else:
            result = tag_manager.execute(name, token, self.context, block,
                                         self, self.loader)
            if callable(result):
                return result
            else:
                return lambda context: ''

    def parse(self, text):
        """Parse template string and create appropriate command list into this
        template instance
        """
        current = Template.TEXT
        token = ''
        cmds = self.commands
        cmd_stack = deque()
        tag_stack = deque()
        token_stack = deque()
        for char in text:
            if current == Template.TEXT:
                if char == '{':
                    current = Template.TOKEN
                    if len(token) > 0:
                        cmds.append(Template.constant(token))
                        token = ''
                else:
                    token += str(char)
            elif current == Template.TOKEN:
                if char == '{':
                    current = Template.ECHO
                elif char == '%':
                    current = Template.TAG
                else:
                    current = Template.TEXT
                    token = '{' + str(char)
            elif current == Template.ECHO or current == Template.FILTER:
                if char == '}':
                    if len(token) > 0:
                        token = token.strip()
                        if current == Template.ECHO:
                            cmd = Template.variable(token)
                        else:
                            cmd = Template.filter(token)
                        cmds.append(cmd)
                        token = ''
                    current = Template.CLOSE
                elif char == '|':
                    current = Template.FILTER
                    token += str(char)
                else:
                    token += str(char)
            elif current == Template.TAG:
                if char == '%':
                    current = Template.CLOSE
                    token = token.strip()
                    name = token.split(' ', 1)[0]
                    if name.startswith('end'):
                        name = name[3:]
                        tag = tag_stack.pop()
                        # Close block
                        if name == tag:
                            block = cmds
                            cmds = cmd_stack.pop()
                            token = token_stack.pop()
                            cmds.append(self.tag(name, token, block))
                        else:
                            raise Exception(
                                "Invalid closing tag: 'end%s' except 'end%s'" %
                                (name, tag))
                    else:
                        if ' ' in token:
                            token = token.split(' ', 1)[1]
                        else:
                            token = ''
                        if tag_manager.is_block_tag(name):
                            cmd_stack.append(cmds)
                            tag_stack.append(name)
                            token_stack.append(token)
                            cmds = []
                        else:
                            cmds.append(self.tag(name, token, ()))
                    token = ''
                else:
                    token += str(char)
            elif current == Template.CLOSE:
                if char == '}':
                    current = Template.TEXT
                else:
                    raise Exception('Wrong template syntax')
            else:
                raise Exception('Wrong template syntax')
        # Check stack length - detect unclosed tags
        if len(cmd_stack) > 0:
            raise Exception('Unexpected end of input - not all tags closed')
        # Last value
        if len(token) > 0:
            cmds.append(Template.constant(token))
        self.commands = cmds

    def execute(self, context):
        """Execute all commands on a specified context

        Arguments:
            context: dict contains varibles
        Returns:
            string contains the whole result
        """
        result = StringIO()
        for cmd in self.commands:
            result.write(cmd(context))
        value = result.getvalue()
        result.close()
        return value

    def __call__(self, context):
        """Alias for execute()
        """
        return self.execute(context)

    def partial(self, context, name='', key_args=()):
        """Execute all commands on a specified context and cache the result as
        another template ready for execution

        Arguments:
            context: dict contains variables
        Returns:
            another template contains the result
        """
        result = Template(loader=self.loader, name=name)
        buffer = StringIO()
        for cmd in self.commands:
            try:
                buffer.write(cmd(context))
            except Exception:
                value = buffer.getvalue()
                if len(value) > 0:
                    result.commands.append(Template.constant(value))
                result.commands.append(cmd)
                buffer.close()
                buffer = StringIO()
        value = buffer.getvalue()
        buffer.close()
        if len(value) > 0:
            result.commands.append(Template.constant(value))
        return result



""" Pakage provides method to working with templates """

from .loaders import TemplateLoader as Loader
from .template import Template as BaseTemplate

from .templatetags import *

TemplateLoader = Loader
Template = BaseTemplate



"""Package provides template filters management
"""


class FilterManager(object):
    """Class used for filters manipulations
    """
    __slots__ = ('apply', 'filters', 'is_filter_exists', )

    def __init__(self):
        """Create new tag managet instance
        """
        super(FilterManager, self).__init__()
        self.filters = {}

    def is_filter_exists(self, name):
        """Check is filter exists
        """
        if name not in self.filters:
            raise Exception("Filter '%s' is not registered" % name)
        return self.filters[name]

    def register(self, filter):
        '''Register filter in manager
        '''
        self.filters[filter.__name__] = filter

    def apply(self, filter, value, args, arg_types, context):
        '''Apply filter to values
        '''
        filter_func = self.is_filter_exists(filter)
        new_args = []
        i = 0
        while i < len(args):
            new_args.append(arg_types[i] and args[i] or context[args[i]])
            i += 1
        return filter_func(value, *new_args)

filter_manager = FilterManager()



#!/usr/bin/env python
"""
Lighty-template
~~~~~~~~~~~~~~~

Lighty-template is very simple template engine for python (python.org). 
Template syntax looks like django-template or jinja2 template. But template 
engine code is easier and gives a way to write all needed tags without any 
hacks. 

Now it does not include all features django-template or jinja2 supports, but 
I'll try to fix it as soon as possible.

Features:
---------

- Stupid simple syntax almost compatible with django-template.
- Pure python.
- Supports both Python 2 (checked with 2.7.2) and Python 3 (checked with 3.2.2)
- Fast. From 3 to 10 times faster than django-template and even faster on some
  benchmarks than jinja2 (but in one benchmark 2 times slower).
- Simple and compact code.
- Template filters with multiply arguments.
- Basic template filters included (now just 14 template filters).
- Basic template tags included.
- Simple but powerfull tag declaration - it's easy to create your own block 
  tags with writing single function.
- Custom template tags can modify template on fly.

Example:
--------

Here a small template example:

	<!DOCTYPE html>
	<html>
	<head>
		<title>{{ title }}</title>
		{% block style %}{% endblock %}
		{% block script %}{% endblock %}
	</head>
	<body>
		{% block content %}
		<h1>Hello, world!</h1>
		<p>Some text here</p>
		{% endblock %}
	</body>
	</html>

TODO:
-----

- More default tags (now there is no load, include tags, and if and for tags
  was simplified)
- More default filters (date formatiing, strings saving, etc.)
- Some additional execution optimizations.
- More tests (in progress).
- Documentation.
"""
from distutils.core import setup

setup(
        name='lighty-template',
        version='0.3.1',
        description='Simple template engine for python',
        long_description=__doc__,
        keywords='Template HTML XML', 
        author='Andrey Grygoryev',
        author_email='undeadgrandse@gmail.com',
        license='BSD',
        url='https://github.com/GrAndSE/lighty-template',
        packages=['lighty', 'lighty.templates'],
        platforms="any",
        classifiers=[
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Text Processing :: Markup :: HTML',
        ],
)



""" Package provides template loaders """


class TemplateLoader(object):
    """ Class fot managing templates """

    def __init__(self, settings=None):
        """ Create new template loader """
        super(TemplateLoader, self).__init__()
        # Create new templates dictionary
        self.templates = {}
        # TODO: do something with settings

    def register(self, name, template):
        """ Add loaded or generated template """
        self.templates[name] = template

    def get_template(self, name):
        """ Get template by name """
        if name not in self.templates:
            raise Exception("Template '%s' was not found" % name)
        return self.templates[name]
